"use strict";

var _fetchMock = _interopRequireDefault(require("fetch-mock"));

var _index = require("../../index");

var _buildStore = _interopRequireDefault(require("../helpers/buildStore"));

var _EmptyKey = _interopRequireDefault(require("../../constants/EmptyKey"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

describe('Generating key:', function () {
  describe('when the urlOnlyParams option is not used', function () {
    beforeAll(function () {
      var _resources = (0, _index.resources)({
        name: 'users',
        url: 'http://test.com/users/:id?',
        keyBy: 'id'
      }, {
        fetchList: true
      }),
          reducers = _resources.reducers,
          fetchUsers = _resources.actionCreators.fetchList;

      this.fetchUsers = fetchUsers;
      this.reducers = reducers;
    });
    describe('when no value is provided', function () {
      beforeAll(function () {
        _fetchMock["default"].get('http://test.com/users', {
          body: [{
            id: 1,
            username: 'Bob'
          }],
          status: 200
        });

        this.store = (0, _buildStore["default"])({
          users: _index.RESOURCES
        }, {
          users: this.reducers
        });
      });
      afterAll(function () {
        _fetchMock["default"].restore();

        this.store = null;
      });
      it('then uses an empty string as the key', function () {
        var _this = this;

        return this.store.dispatch(this.fetchUsers()).then(function () {
          var list = _this.store.getState().users.lists[_EmptyKey["default"]];

          expect(list.positions).toEqual([1]);
          expect(list.status.type).toEqual(_index.SUCCESS);
        });
      });
    });
    describe('when an object is provided with only the id', function () {
      beforeAll(function () {
        _fetchMock["default"].get('http://test.com/users/newest', {
          body: [{
            id: 1,
            username: 'Bob'
          }],
          status: 200
        });

        this.store = (0, _buildStore["default"])({
          users: _index.RESOURCES
        }, {
          users: this.reducers
        });
      });
      afterAll(function () {
        _fetchMock["default"].restore();

        this.store = null;
      });
      it('then uses the id as a key', function () {
        var _this2 = this;

        return this.store.dispatch(this.fetchUsers({
          id: 'newest'
        })).then(function () {
          var list = _this2.store.getState().users.lists['id=newest'];

          expect(list.positions).toEqual([1]);
          expect(list.status.type).toEqual(_index.SUCCESS);
        });
      });
    });
    describe('when an object is provided without the id attribute', function () {
      beforeAll(function () {
        _fetchMock["default"].get('http://test.com/users?order=newest&page=1', {
          body: [{
            id: 1,
            username: 'Bob'
          }],
          status: 200
        });

        this.store = (0, _buildStore["default"])({
          users: _index.RESOURCES
        }, {
          users: this.reducers
        });
      });
      afterAll(function () {
        _fetchMock["default"].restore();

        this.store = null;
      });
      it('then uses a serialised version of the object as a key', function () {
        var _this3 = this;

        return this.store.dispatch(this.fetchUsers({
          order: 'newest',
          page: 1
        })).then(function () {
          var list = _this3.store.getState().users.lists['order=newest.page=1'];

          expect(list.positions).toEqual([1]);
          expect(list.status.type).toEqual(_index.SUCCESS);
        });
      });
    });
  });
  describe('when the urlParamsOnly is used', function () {
    beforeAll(function () {
      var _resources2 = (0, _index.resources)({
        name: 'users',
        url: 'http://test.com/users/:id?',
        keyBy: 'id',
        urlOnlyParams: ['page']
      }, {
        fetchList: true
      }),
          reducers = _resources2.reducers,
          fetchUsers = _resources2.actionCreators.fetchList;

      this.reducers = reducers;
      this.fetchUsers = fetchUsers;
    });
    describe('when an object with the attribute mentioned in urlOnlyParams is used', function () {
      beforeAll(function () {
        _fetchMock["default"].get('http://test.com/users?order=newest&page=1', {
          body: [{
            id: 1,
            username: 'Bob'
          }],
          status: 200
        });

        this.store = (0, _buildStore["default"])({
          users: _index.RESOURCES
        }, {
          users: this.reducers
        });
      });
      afterAll(function () {
        _fetchMock["default"].restore();

        this.store = null;
      });
      it('then does NOT use that attribute in the serialised object used for the key', function () {
        var _this4 = this;

        return this.store.dispatch(this.fetchUsers({
          order: 'newest',
          page: 1
        })).then(function () {
          var list = _this4.store.getState().users.lists['order=newest'];

          expect(list.positions).toEqual([1]);
          expect(list.status.type).toEqual(_index.SUCCESS);
        });
      });
    });
    describe('when an object with the attribute mentioned in urlOnlyParams is used', function () {
      beforeAll(function () {
        _fetchMock["default"].get('http://test.com/users?order=newest&page=1', {
          body: [{
            id: 1,
            username: 'Bob'
          }],
          status: 200
        });

        _fetchMock["default"].get('http://test.com/users?order=newest&page=2', {
          body: [{
            id: 2,
            username: 'Jane'
          }],
          status: 200
        });

        this.store = (0, _buildStore["default"])({
          users: _index.RESOURCES
        }, {
          users: this.reducers
        });
      });
      afterAll(function () {
        _fetchMock["default"].restore();

        this.store = null;
      });
      it('then overrides the results for the fetchList action', function () {
        var _this5 = this;

        return this.store.dispatch(this.fetchUsers({
          order: 'newest',
          page: 1
        })).then(function () {
          return _this5.store.dispatch(_this5.fetchUsers({
            order: 'newest',
            page: 2
          })).then(function () {
            _this5.users = _this5.store.getState().users;
            expect(_this5.users.items['1'].values).toEqual({
              id: 1,
              username: 'Bob'
            });
            expect(_this5.users.items['1'].status.type).toEqual(_index.SUCCESS);
            expect(_this5.users.items['2'].values).toEqual({
              id: 2,
              username: 'Jane'
            });
            expect(_this5.users.items['2'].status.type).toEqual(_index.SUCCESS);
            expect(_this5.users.lists['order=newest'].positions).toEqual([2]);
          });
        });
      });
    });
  });
});