"use strict";

var _fetchMock = _interopRequireDefault(require("fetch-mock"));

var _index = require("../../../index");

var _nop = _interopRequireDefault(require("../../../utils/function/nop"));

var _resourceAssertions = require("../../helpers/resourceAssertions");

var _EmptyKey = _interopRequireDefault(require("../../../constants/EmptyKey"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var RESOURCE_NAME = 'users';
describe('DestroyItem reducer:', function () {
  beforeAll(function () {
    var _resource = (0, _index.resource)({
      name: 'users',
      url: 'http://test.com/users',
      keyBy: 'id'
    }, {
      destroyItem: true
    }),
        reducers = _resource.reducers,
        destroyUser = _resource.actionCreators.destroyItem;

    this.reducers = reducers;
    this.destroyUser = destroyUser;
  });
  describe('Given the resource item is NOT in the store', function () {
    expectToHandleSuccessAndFailedRequests({
      items: {},
      lists: {},
      selectionMap: {},
      newItemKey: null
    }, "Redux and the REST: DESTROY_USER's key '".concat(_EmptyKey["default"], "' did not match any items in the store. (Destroy request was still sent to the server.)"));
  });
  describe('Given the resource item is in the store with a status of NEW', function () {
    expectToHandleSuccessAndFailedRequests(getInitialState(_index.NEW), "Redux and the REST: DESTROY_USER's key '".concat(_EmptyKey["default"], "' matched a new item. Use clearNewItem() to clear items that haven't been saved to the server. (Destroy request was still sent to the server.)"));
  });
  describe('Given a resources item is in the store with a status of DESTROYING', function () {
    expectToHandleSuccessAndFailedRequests(getInitialState(_index.DESTROYING), "Redux and the REST: DESTROY_USER's key '".concat(_EmptyKey["default"], "' matched a new item that has a pending DESTROY action. (Duplicate destroyItem request was still sent to the server.)"));
  });
  describe('Given a resources item is in the store with a status of EDITING', function () {
    expectToHandleSuccessAndFailedRequestsForExistingResourceItem(getInitialState(_index.EDITING));
  });
  describe('Given a resources item is in the store with a status of CREATING', function () {
    expectToHandleSuccessAndFailedRequestsForExistingResourceItem(getInitialState(_index.CREATING));
  });
  describe('Given a resources item is in the store with a status of UPDATING', function () {
    expectToHandleSuccessAndFailedRequestsForExistingResourceItem(getInitialState(_index.UPDATING));
  });
  describe('Given a resources item is in the store with a status of SUCCESS', function () {
    expectToHandleSuccessAndFailedRequestsForExistingResourceItem(getInitialState(_index.SUCCESS));
  });
  describe('Given a resources item is in the store with a status of ERROR', function () {
    expectToHandleSuccessAndFailedRequestsForExistingResourceItem(getInitialState(_index.ERROR));
  });
  describe('Given a destroyItem action that will succeed with a response that specifies \'errors\' at the top level', function () {
    describe('when the request has completed', function () {
      expectToHandleErrorsCorrectly(200);
    });
  });
  describe('Given a destroyItem action that will fail with a response that specifies \'errors\' at the top level', function () {
    describe('when the request has completed', function () {
      expectToHandleErrorsCorrectly(404);
    });
  });

  function expectToHandleSuccessAndFailedRequests(initialState) {
    var warning = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
    beforeAll(function () {
      spyOn(console, 'warn');
    });
    describe('before the request has completed', function () {
      beforeAll(function () {
        setUpBeforeRequest(this, initialState);
      });
      afterAll(function () {
        tearDown(this);
      });
      it('then warns about the missing resource', function () {
        expect(console.warn).toHaveBeenCalledWith(warning);
      });
      it('then creates a new item and sets its status to DESTROYING', function () {
        (0, _resourceAssertions.expectToChangeResourceItemStatusTo)(this, RESOURCE_NAME, 'type', _index.DESTROYING);
      });
    });
    describe('when the API request succeeds', function () {
      var _this = this;

      beforeAll(function () {
        setUpAfterRequestSuccess(this, initialState);
      });
      afterAll(function () {
        return tearDown(_this);
      });
      it('then removes the item', function () {
        expect((0, _resourceAssertions.resourcesDefinition)(this, RESOURCE_NAME).items).toEqual({});
      });
    });
    describe('when the API request errors', function () {
      beforeAll(function () {
        setUpAfterRequestFailure(this, initialState);
      });
      afterAll(function () {
        tearDown(this);
      });
      expectToRecordDestroyError(404);
    });
  }

  function expectToRecordDestroyError(status) {
    it('then updates the items\'s status to DESTROY_ERROR', function () {
      (0, _resourceAssertions.expectToChangeResourceItemStatusTo)(this, RESOURCE_NAME, 'type', _index.DESTROY_ERROR);
    });
    it('then sets the syncedAt attribute', function () {
      (0, _resourceAssertions.expectToChangeResourceItemStatusErrorOccurredAtToBeSet)(this, RESOURCE_NAME);
    });
    it('then updates the item\'s status httpCode', function () {
      (0, _resourceAssertions.expectToChangeResourceItemStatusTo)(this, RESOURCE_NAME, 'httpCode', status);
    });
  }

  function expectToHandleErrorsCorrectly(status) {
    beforeAll(function () {
      setUpAfterRequestFailure(this, getInitialState(_index.SUCCESS), {
        body: {
          errors: [{
            message: 'Not Found'
          }]
        },
        status: status
      });
    });
    afterAll(function () {
      tearDown(this);
    });
    expectToRecordDestroyError(status);
    it('then sets the item\'s errors', function () {
      (0, _resourceAssertions.expectToChangeResourceItemStatusTo)(this, RESOURCE_NAME, 'error', {
        message: 'Not Found'
      });
      (0, _resourceAssertions.expectToChangeResourceItemStatusTo)(this, RESOURCE_NAME, 'errors', [{
        message: 'Not Found'
      }]);
    });
  }

  function expectToHandleSuccessAndFailedRequestsForExistingResourceItem(initialState) {
    describe('before the request has completed', function () {
      var _this2 = this;

      beforeAll(function () {
        setUpBeforeRequest(this, initialState);
      });
      afterAll(function () {
        return tearDown(_this2);
      });
      it('then creates a new item and sets its status to DESTROYING', function () {
        (0, _resourceAssertions.expectToChangeResourceItemStatusTo)(this, RESOURCE_NAME, 'type', _index.DESTROYING);
      });
      it('then does NOT remove the item from the selectionMap', function () {
        (0, _resourceAssertions.expectToNotChangeSelectionMap)(this, RESOURCE_NAME);
      });
      it('then does NOT clear the item from thew newItemKey', function () {
        (0, _resourceAssertions.expectToNotChangeNewItemKey)(this, RESOURCE_NAME);
      });
    });
    describe('when the API request succeeds', function () {
      var _this3 = this;

      beforeAll(function () {
        setUpAfterRequestSuccess(this, initialState);
      });
      afterAll(function () {
        return tearDown(_this3);
      });
      it('then removes the item', function () {
        expect((0, _resourceAssertions.resourcesDefinition)(this, RESOURCE_NAME).items).toEqual({});
      });
      it('then removes the deleted item from the selectionMap', function () {
        (0, _resourceAssertions.expectToChangeSelectionMapTo)(this, RESOURCE_NAME, {});
      });
      it('then removes the deleted item from the newItemKey', function () {
        (0, _resourceAssertions.expectToChangeNewItemKeyTo)(this, RESOURCE_NAME, null);
      });
    });
    describe('when the API request errors', function () {
      beforeAll(function () {
        setUpAfterRequestFailure(this, initialState);
      });
      afterAll(function () {
        tearDown(this);
      });
      expectToRecordDestroyError(404);
      it('then does NOT clear the item from thew newItemKey', function () {
        (0, _resourceAssertions.expectToNotChangeNewItemKey)(this, RESOURCE_NAME);
      });
      it('then updates the item\'s status httpCode', function () {
        (0, _resourceAssertions.expectToChangeResourceItemStatusTo)(this, RESOURCE_NAME, 'httpCode', 404);
      });
    });
  }

  function getInitialState(statusType) {
    return {
      items: _defineProperty({}, _EmptyKey["default"], {
        values: {
          username: 'Bob'
        },
        status: {
          type: statusType
        }
      }),
      lists: {},
      selectionMap: {},
      newItemKey: null
    };
  }

  function setUpBeforeRequest(context, initialState) {
    _fetchMock["default"]["delete"]('http://test.com/users', new Promise(_nop["default"]));

    setupState(context, initialState);
  }

  function setUpAfterRequestSuccess(context, initialState) {
    var actionCreatorOptions = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _fetchMock["default"]["delete"]('http://test.com/users', {
      body: {}
    });

    setupState(context, initialState, actionCreatorOptions);
  }

  function setUpAfterRequestFailure(context, initialState) {
    var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {
      body: {
        error: 'Not Found'
      },
      status: 404
    };

    _fetchMock["default"]["delete"]('http://test.com/users', options);

    setupState(context, initialState);
  }

  function setupState(context, initialState) {
    (0, _resourceAssertions.setupInitialState)(context, RESOURCE_NAME, initialState);
    context.store.dispatch(context.destroyUser());
  }

  function tearDown(context) {
    _fetchMock["default"].restore();

    context.store = null;
  }
});